"""Plan service

@author Rory Byrne <rory@rory.bio>
"""
import os
import tempfile
import time
from subprocess import call
from typing import List

from git_plan.exceptions import PlanEmpty
from git_plan.model.commit import Commit, CommitMessage
from git_plan.model.repository import Repository
from git_plan.service.git import GitService
from git_plan.util.unix import is_installed


class PlanService:
    """Manage the user's plans"""

    def __init__(self, plan_template: str, edit_template: str, git_service: GitService):
        assert edit_template, "Edit template missing"
        assert plan_template, "Commit template missing"
        self._plan_template = plan_template
        self._edit_template = edit_template
        self._git_service = git_service

    def add_commit(self, repository: Repository) -> Commit:
        """Create a plan in the given directory"""
        commit_id = str(int(time.time()))
        commit = self._create_commit(repository, commit_id)
        commit.save()
        return commit

    def edit_commit(self, commit: Commit):
        """Update the plan in the given directory"""
        template = self._edit_template \
            .replace('%headline%', commit.message.headline) \
            .replace('%body%', commit.message.body)

        new_message = self._prompt_user_for_plan(initial=template)
        commit.message = new_message
        commit.updated_at = int(time.time())
        commit.save()

    @staticmethod
    def delete_commit(commit: Commit):
        """Delete the chosen commit"""
        if not commit.path.exists():
            raise RuntimeError(f'Commit not found: {commit}')

        commit.path.unlink()  # Deletes the file

    @staticmethod
    def has_commits(repository: Repository) -> bool:
        """Check if a plan already exists in the given directory"""
        return any(repository.plan_files_dir.iterdir())  # False if it cannot iterate at least once

    def get_commits(self, repository: Repository, branch: str = None) -> List[Commit]:
        """Print the status of the plan

        Raises:
            RuntimeError:   Commit file not found
        """
        return self._fetch_commits(repository, branch=branch)

    # Private #############

    def _create_commit(self, repository: Repository, commit_id: str) -> Commit:
        message = self._prompt_user_for_plan()
        if not message or message.headline == '':
            raise RuntimeError("Invalid commit plan. Please include at least a headline.")

        branch = self._git_service.get_current_branch()
        created_at: float = time.time()
        updated_at: float = created_at
        commit = Commit(repository, commit_id, branch, int(created_at), int(updated_at))
        commit.message = message
        return commit

    def _prompt_user_for_plan(self, initial: str = None) -> CommitMessage:
        if not initial:
            initial = self._plan_template

        editor = self._git_service.get_configured_editor()
        if not editor:
            editor = os.environ.get("EDITOR", "vim")
        if not is_installed(editor):
            raise RuntimeError("Couldn't find an editor installed on your system.")

        with tempfile.NamedTemporaryFile(suffix=".tmp", mode='r+') as file:
            file.write(initial)
            file.flush()
            call([editor, file.name])

            file.seek(0)
            message_lines = file.readlines()
            processed_input = self._post_process_commit(message_lines)

            return CommitMessage.from_string(processed_input)

    def _fetch_commits(self, repository: Repository, branch: str = None) -> List['Commit']:
        if not self.has_commits(repository):
            return []

        commit_files = repository.plan_files_dir.iterdir()
        commits: List[Commit] = [Commit.from_file(f, repository) for f in commit_files if f.is_file()]
        if branch:
            # .strip() for backawrds compatibility
            commits = [commit for commit in commits if commit.branch.strip() == branch]

        return commits

    @staticmethod
    def _post_process_commit(lines: List[str]):
        lines = [line.strip() for line in lines if not line.startswith('#') or line == '\n']
        if not lines or len(lines) == 0:
            raise PlanEmpty()

        headline = lines[0].strip()
        body = '\n'.join(lines[1:]).strip()

        return ''.join([headline, '\n', '\n', body])



"""UI service

Author: Rory Byrne <rory@rory.bio>
"""
import datetime as dt
from typing import List

import humanize
import inquirer
from rich import print as rich_print

from git_plan.model.commit import Commit


class UIService:
    """Functionality for rendering in the terminal"""

    @staticmethod
    def choose_commit(commits: List[Commit], message: str):
        """Choose from a set of tasks"""
        if len(commits) == 0:
            raise RuntimeError("Cannot choose from an empty list of commits.")

        if len(commits) == 1:
            return commits[0]

        options = [
            inquirer.List(
                'commit',
                message=message,
                choices=[(c.message.headline, c) for c in commits]
            )
        ]

        answer = inquirer.prompt(options)

        return answer['commit']

    @staticmethod
    def confirm(message: str) -> bool:
        """Ask the user for confirmation"""
        key = 'confirm'
        questions = [inquirer.Confirm(key, message=message)]

        answers = inquirer.prompt(questions)

        return answers[key]

    def bold(self, message: str):
        """Print a bolded message"""
        self.print(f'[bold]{message}[/bold]')

    @staticmethod
    def print(message: str):
        """Print a message to the terminal"""
        rich_print(message)

    def render_commits(self, commits: List[Commit], headline_only=True):
        """Renders a list of commits in pretty print"""
        for idx, commit in enumerate(commits):
            self._render_commit(commit, str(idx + 1), headline_only)
            if idx < len(commits) - 1:
                print('')

    def _render_commit(self, commit: Commit, tag: str, headline_only):
        if not commit.updated_at:
            raise ValueError("Invalid commit: no updated_at field found.")

        time_display = f"[dim]({humanize.naturaltime(dt.datetime.fromtimestamp(commit.updated_at))})[/dim]"
        if headline_only:
            self.print(f'[bold][{tag}][/bold] {commit.message.headline} {time_display}')
        else:
            self.print(f'[bold][[magenta]{tag}[/magenta]][/bold] on {commit.branch} {time_display}\n')
            self.print(f'    {commit.message.headline}\n')
            body_lines = commit.message.body.split('\n')
            for line in body_lines:
                self.print(f'    {line}')



"""CLI entry point

Author: Rory Byrne <rory@rory.bio>
"""
import sys
from pathlib import Path
from typing import List

from dependency_injector.wiring import inject, Provide

from git_plan.cli.cli import CLI
from git_plan.conf import Settings
from git_plan.containers import Application
from git_plan.exceptions import NotAGitRepository, NotInitialized, GitPlanException

HOME = str(Path.home())


def main():
    """Entrypoint"""
    settings = Settings.load()

    app = Application()
    app.config.from_dict(settings)
    app.wire(modules=[sys.modules[__name__]])

    args = sys.argv[1:]  # Might be []

    try:
        launch_cli(args)
    except NotInitialized:
        print("Git plan is not initialized.\n\tPlease run `git plan init`")
    except NotAGitRepository:
        print("You are not in a git repository (no .git/ directory found).")
    except GitPlanException as exc:
        print("git plan encountered an error.")
        print("Please open an issue at https://github.com/synek/git-plan and let us know.\n")
        print(f"The error message was '{str(exc)}'")


@inject
def launch_cli(args: List[str], cli: CLI = Provide[Application.cli]):
    """Inject dependency tree and run the CLI"""
    cli.parse(args)


if __name__ == "__main__":
    main()



"""Delete command

Author: Rory Byrne <rory@rory.bio>
"""
from typing import Any

from git_plan.cli.commands.command import Command
from git_plan.service.plan import PlanService
from git_plan.util.decorators import requires_initialized, requires_git_repository


@requires_initialized
@requires_git_repository
class Delete(Command):
    """Delete an existing commit"""

    subcommand = 'delete'

    def __init__(self, plan_service: PlanService, **kwargs):
        super().__init__(**kwargs)
        assert plan_service, "Plan service not injected"
        self._plan_service = plan_service

    def command(self, **kwargs):
        """Create a new commit"""
        commits = self._plan_service.get_commits(self._repository)
        if not commits:
            self._ui.bold('No commits found.')
            return

        chosen_commit = self._ui.choose_commit(commits, 'Which plan do you want to delete?')

        self._ui.bold(f'{chosen_commit.message.headline}\n')
        confirm_msg = 'Are you sure you want to delete this commit?'
        if not self._ui.confirm(confirm_msg):
            self._ui.bold("Stopped.")
            return

        self._plan_service.delete_commit(chosen_commit)
        self._ui.bold('Deleted.')

    def register_subparser(self, subparsers: Any):
        subparsers.add_parser(Delete.subcommand, help='Delete a planned commit.')



"""Base class for Commands

@author Rory Byrne <rory@rory.bio>
"""
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Any

from git_plan.model.repository import Repository
from git_plan.service.ui import UIService

if TYPE_CHECKING:
    from git_plan.cli.cli import CLI
else:
    CLI = Any


class Command(ABC):
    """Base for all Commands"""

    subcommand: str

    def __init__(self, repository: Optional[Repository], ui_service: UIService):
        self._cli: Optional[CLI] = None
        self._repository: Optional[Repository] = repository
        self._ui = ui_service

    def run(self, context: dict):
        """Run the command"""
        self.command(**context)

    def set_cli(self, cli: 'CLI'):
        """Set the CLI reference, so that a command can trigger other commands"""
        assert cli, "Cannot set CLI: None"
        self._cli = cli

    @abstractmethod
    def register_subparser(self, subparsers: Any):
        """Allows a command to define its own args"""
        raise NotImplementedError()

    @abstractmethod
    def command(self, **kwargs):
        """The implementation of the command's functionality"""
        raise NotImplementedError()



"""Commit command

Author: Rory Byrne <rory@rory.bio>
"""
from typing import Any

from git_plan.cli.commands.command import Command
from git_plan.exceptions import CommitAbandoned
from git_plan.service.git import GitService
from git_plan.service.plan import PlanService
from git_plan.util.decorators import requires_initialized, requires_git_repository


@requires_initialized
@requires_git_repository
class Commit(Command):
    """Commit a planned commit"""

    subcommand = 'commit'

    def __init__(self, plan_service: PlanService, git_service: GitService, **kwargs):
        super().__init__(**kwargs)
        assert plan_service, "Plan service not injected"
        assert git_service, "Git service not injected"
        self._plan_service = plan_service
        self._git_service = git_service

    def command(self, **kwargs):
        """Create a new commit"""
        commits = self._plan_service.get_commits(self._repository)
        if not commits:
            print("No commits planned.")
            return

        if not self._git_service.has_staged_files():
            print("No staged files.")
            return

        chosen_commit = self._ui.choose_commit(commits, 'Which plan do you want to commit?')
        try:
            self._git_service.commit(chosen_commit)
            self._plan_service.delete_commit(chosen_commit)
        except CommitAbandoned:
            print("Commit abandoned.")

    def register_subparser(self, subparsers: Any):
        subparsers.add_parser(Commit.subcommand, help='Commit a plan.')



"""GitPlanExceptions

@author Rory Byrne <rory@rory.bio
"""


class GitPlanException(Exception):
    """Base exception for git plan"""


class CommandNotFound(GitPlanException):
    """Command doesn't exist"""


class NotInitialized(GitPlanException):
    """The repo was not initialized"""


class AlreadyInitialized(GitPlanException):
    """The repo is already initialized"""


class CommitAbandoned(GitPlanException):
    """The user abandoned their commit"""


class PlanEmpty(GitPlanException):
    """The commit plan was empty"""


class GitException(GitPlanException):
    """Base class for git-related exceptions"""


class NotAGitRepository(GitException):
    """The current repo is not a git repository"""



"""Add command

@author Rory Byrne <rory@rory.bio>
"""
from typing import Any

from git_plan.cli.commands.command import Command
from git_plan.exceptions import PlanEmpty
from git_plan.service.plan import PlanService
from git_plan.util.decorators import requires_initialized, requires_git_repository


@requires_initialized
@requires_git_repository
class Add(Command):
    """Add a new commit"""

    subcommand = 'add'

    def __init__(self, plan_service: PlanService, **kwargs):
        super().__init__(**kwargs)
        assert plan_service, "Plan service not injected"
        self._plan_service = plan_service

    def command(self, **kwargs):
        """Create a new commit"""
        try:
            self._plan_service.add_commit(self._repository)
        except PlanEmpty:
            self._ui.bold('Plan empty, abandoning.')

    def register_subparser(self, subparsers: Any):
        subparsers.add_parser(Add.subcommand, help='Add a new commit plan.')



"""Init command

@author Rory Byrne <rory@rory.bio>
"""
from typing import Any

from git_plan.cli.commands.command import Command
from git_plan.exceptions import AlreadyInitialized
from git_plan.service.repository import RepositoryService
from git_plan.util.decorators import requires_git_repository


@requires_git_repository
class Init(Command):
    """Initialize git plan in the directory."""

    subcommand = 'init'

    def __init__(self, repository_service: RepositoryService, **kwargs):
        super().__init__(**kwargs)
        assert repository_service, "Repository service not injected"
        self._repo_service = repository_service

    def command(self, **kwargs):
        """Initialize the repository if it is not already initialized"""
        if not self._repository:
            self._ui.bold("Not in a git repository.")
            return

        try:
            self._repo_service.initialize(self._repository)
            self._ui.print(f"Initialized git plan in [bold]{self._repository.plan_dir}[/bold]")
        except AlreadyInitialized:
            self._ui.bold("Git plan is already initialized.")

    def register_subparser(self, subparsers: Any):
        subparsers.add_parser(Init.subcommand, help='Initialize git plan.')



"""Git service

Author: Rory Byrne <rory@rory.bio>
"""
import subprocess
from typing import Optional

from git_plan.exceptions import CommitAbandoned, GitException
from git_plan.model.commit import Commit
from git_plan.util import unix
from git_plan.util.decorators import requires_git_repository


class GitService:
    """Interface to git functionality"""

    COMMIT = 'git commit -e -m'
    HAS_STAGED = 'git diff --staged --quiet'
    GET_BRANCH = 'git branch --show-current'
    CONFIGURED_GIT_EDITOR = 'git config --global core.editor'

    def __init__(self):
        pass

    @requires_git_repository
    def commit(self, commit: Commit):
        """Runs git commit with the given commit-plan as a template"""
        cmd = self.COMMIT.split(' ')
        cmd.append(str(commit.message))

        try:
            unix.run_command(cmd, capture_output=False)
        except subprocess.CalledProcessError as e:
            raise CommitAbandoned() from e

    @requires_git_repository
    def has_staged_files(self) -> bool:
        """Returns True if there are staged files, and False otherwise"""
        cmd = self.HAS_STAGED.split(' ')

        try:
            unix.run_command(cmd)
            return False
        except subprocess.CalledProcessError:
            return True

    @requires_git_repository
    def get_current_branch(self):
        """Gets the current branch via git"""
        cmd = self.GET_BRANCH.split(' ')

        try:
            branch = unix.run_command(cmd)
            if not branch or branch == '':
                raise GitException(f'Invalid branch: "{branch}"')

            return branch.strip()
        except subprocess.CalledProcessError as e:
            raise GitException('Failed to get current git branch') from e

    @requires_git_repository
    def get_configured_editor(self) -> Optional[str]:
        """Gets the editor configured for git"""
        cmd = self.CONFIGURED_GIT_EDITOR.split(' ')

        try:
            editor = unix.run_command(cmd)
            return editor.strip() if editor else None
        except subprocess.CalledProcessError:
            return None



"""Repository service"""

from git_plan.exceptions import AlreadyInitialized
from git_plan.model.repository import Repository


class RepositoryService:  # pylint: disable=too-few-public-methods
    """Administrative functionality for repositories"""

    @staticmethod
    def initialize(repo: Repository):
        """Create the plan directory"""
        if repo.plan_dir.exists():
            raise AlreadyInitialized()

        repo.plan_dir.mkdir()
        repo.plan_files_dir.mkdir()



from pathlib import Path

VERSION = {}
try:
    with open(Path(__file__).parent / "_version", "r") as file:
        VERSION["__version__"] = file.read().strip()
except FileNotFoundError:
    VERSION["__version__"] = "dev"

__version__ = VERSION["__version__"]



"""Unix utilities

Author Rory Byrne <rory@rory.bio>
"""
from shutil import which
import subprocess
from typing import List, Optional


def is_installed(name: str) -> bool:
    """Check whether `name` is on PATH and marked as executable."""
    return which(name) is not None


def run_command(cmd: List[str], capture_output: bool = True) -> Optional[str]:
    """Run a shell command"""
    result = subprocess.run(cmd, capture_output=capture_output, check=True)
    if result.stdout:
        return result.stdout.decode()

    return None



"""Commit model

Author: Rory Byrne <rory@rory.bio>
"""
import json
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from git_plan.exceptions import GitPlanException, NotInitialized
from git_plan.model.repository import Repository

COMMIT_FILE_EXT = '.txt'


@dataclass
class CommitMessage:
    """The message of a commit"""
    headline: str
    body: str

    def __str__(self):
        return f'''{self.headline}

{self.body}
        '''

    @classmethod
    def from_file(cls, path: Path):
        """Load a commit's message from a file"""
        try:
            with open(path, 'r') as file:
                commit_data = json.load(file)

            return CommitMessage(**commit_data['message'])
        except Exception as exc:
            raise GitPlanException(f'Failed to load commit from disk: {path}') from exc

    @classmethod
    def from_string(cls, string: str):
        """Construct a CommitMessage from a well-formatted string"""
        components = string.split('\n\n')
        if len(components) == 2:
            return CommitMessage(components[0], components[1])

        if len(components) > 2:
            return CommitMessage(components[0], '\n\n'.join(components[1:]))

        raise GitPlanException(f'Could not parse commit message from string: "{string}"')


@dataclass
class Commit:
    """Represents a planned commit"""
    repository: Repository
    id: str
    branch: str
    created_at: int
    updated_at: int
    _message: Optional[CommitMessage] = field(init=False, default=None)

    @property
    def filename(self):
        """The filename where this plan is stored"""
        return f'commit-{self.id}'

    @property
    def path(self) -> Path:
        """The absolute path where this plan is stored"""
        return Path(self.repository.plan_files_dir / self.filename).with_suffix(COMMIT_FILE_EXT).resolve()

    @property
    def message(self):
        """Returns the commit message, loading it from disk if needed"""
        if not self._message:
            try:
                self.message = CommitMessage.from_file(self.path)
            except RuntimeError as exc:
                raise RuntimeError(f"Commit doesn't exist at location: {self.path}") from exc

        return self._message

    @message.setter
    def message(self, value: CommitMessage):
        self._message = value

    def save(self):
        """Persist the commit to the storage"""
        if not self._message:
            raise RuntimeError("Cannot save a commit with no message.")

        if not self.repository.is_initialized():
            raise NotInitialized()

        if not self.created_at:
            raise ValueError("Missing created_at")

        if not self.created_at:
            raise ValueError("Missing updated_at")

        commit_dict = {
            "branch": self.branch.strip(),
            'message': {
                "headline": self.message.headline.strip(),
                "body": self.message.body.strip()
            },
            "created_at": self.created_at,
            "updated_at": self.updated_at
        }

        with open(self.path, 'w') as file:
            file.write(json.dumps(commit_dict))

    @classmethod
    def from_file(cls, file: Path, repository: Repository) -> "Commit":
        """Load a commit from a file"""
        with open(file, 'r') as fp:
            commit_data = json.load(fp)

        commit_message = CommitMessage(**commit_data['message'])
        commit_id = file.stem.split('-')[1]
        branch = commit_data['branch']
        created_at = commit_data.get('created_at', time.time())
        updated_at = commit_data.get('updated_at', created_at)

        commit = Commit(
            repository,
            commit_id,
            branch,
            int(created_at),
            int(updated_at)
        )
        commit.message = commit_message

        return commit

    def __str__(self):
        if not self._message:
            return super().__str__()
        return self._message.__str__()



"""Defines the dependency injection structure

@author Rory Byrne <rory@rory.bio>
"""
# pylint: disable=no-member
from dependency_injector import containers, providers

from git_plan.cli.cli import CLI
from git_plan.cli.commands.add import Add
from git_plan.cli.commands.commit import Commit
from git_plan.cli.commands.delete import Delete
from git_plan.cli.commands.edit import Edit
from git_plan.cli.commands.init import Init
from git_plan.cli.commands.list import List
from git_plan.model.repository import Repository
from git_plan.service.git import GitService
from git_plan.service.plan import PlanService
from git_plan.service.repository import RepositoryService
from git_plan.service.ui import UIService


class Core(containers.DeclarativeContainer):
    """Global configuration for the system"""
    config = providers.Configuration()
    repository = providers.Singleton(
        Repository.from_working_dir,
        working_dir=config.working_dir
    )


class Services(containers.DeclarativeContainer):
    """Dependency structure for services"""
    config = providers.Configuration()

    git_service = providers.Singleton(GitService)
    plan_service = providers.Singleton(
        PlanService,
        plan_template=config.template.plan,
        edit_template=config.template.edit,
        git_service=git_service,
    )
    repository = providers.Singleton(RepositoryService)
    ui_service = providers.Singleton(UIService)


class Commands(containers.DeclarativeContainer):
    """Dependency structure for Commands"""
    config = providers.Configuration()
    services = providers.DependenciesContainer()
    core = providers.DependenciesContainer()

    list_command = providers.Singleton(
        List,
        plan_service=services.plan_service,
        ui_service=services.ui_service,
        git_service=services.git_service,
        repository=core.repository
    )
    add_command = providers.Singleton(
        Add,
        plan_service=services.plan_service,
        ui_service=services.ui_service,
        repository=core.repository
    )
    edit_command = providers.Singleton(
        Edit,
        ui_service=services.ui_service,
        plan_service=services.plan_service,
        repository=core.repository
    )
    delete_command = providers.Singleton(
        Delete,
        ui_service=services.ui_service,
        plan_service=services.plan_service,
        repository=core.repository
    )
    commit_command = providers.Singleton(
        Commit,
        ui_service=services.ui_service,
        plan_service=services.plan_service,
        git_service=services.git_service,
        repository=core.repository
    )
    init_command = providers.Singleton(
        Init,
        repository_service=services.repository,
        ui_service=services.ui_service,
        repository=core.repository
    )


class Application(containers.DeclarativeContainer):
    """Top-level container for the application"""
    config = providers.Configuration()

    core = providers.Container(
        Core,
        config=config
    )

    services = providers.Container(
        Services,
        config=config,
    )

    commands = providers.Container(
        Commands,
        config=config,
        services=services,
        core=core
    )

    # Entrypoints
    cli = providers.Singleton(
        CLI,
        commands=providers.List(
            commands.list_command,
            commands.add_command,
            commands.edit_command,
            commands.commit_command,
            commands.init_command,
            commands.delete_command
        ),
        plan_service = services.plan_service,
        repository = core.repository
    )



"""Decorators

Author: Rory Byrne <rory@rory.bio>
"""
import subprocess
from functools import wraps
from inspect import isclass

from git_plan.exceptions import NotAGitRepository, NotInitialized
from git_plan.util import unix


def _shell_is_in_git_repository():
    command = 'git rev-parse --is-inside-work-tree'
    cmd = command.split(' ')
    try:
        unix.run_command(cmd)
        return True
    except subprocess.CalledProcessError:
        return False


def requires_git_repository(ref):
    """Raises NotAGitRepository if the check for a git repository fails"""
    if isclass(ref):
        if not hasattr(ref, 'run'):
            raise ValueError("Cannot use @requires_git_repository on this class")
        ref.run = requires_git_repository(ref.run)
        return ref

    @wraps(ref)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, '_repository'):
            if self._repository is None:  # pylint: disable=protected-access
                raise NotAGitRepository()
        else:  # fallback to regular shell command
            if not _shell_is_in_git_repository():
                raise NotAGitRepository()

        return ref(self, *args, **kwargs)

    return wrapper


def requires_initialized(ref):
    """Checks that the repository in the arguments is initialized"""
    if isclass(ref):
        if not hasattr(ref, 'run'):
            raise ValueError("Cannot use @requires_initialized on this class")
        ref.run = requires_initialized(ref.run)
        return ref

    @wraps(ref)
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, '_repository'):
            raise ValueError("Cannot use the @requires_initialized decorator here.")

        if self._repository is None:  # pylint: disable=protected-access
            raise NotAGitRepository()

        is_initialized = self._repository.is_initialized()  # pylint: disable=protected-access

        if is_initialized:
            return ref(self, *args, **kwargs)

        raise NotInitialized()

    return wrapper



"""CLI Entrypoint"""
import argparse
from argparse import Namespace
from typing import Dict, List

import pkg_resources

from git_plan.cli.commands.command import Command
from git_plan.exceptions import CommandNotFound
from git_plan.model.repository import Repository
from git_plan.service.plan import PlanService


class CLI:
    """The command-line entrypoint"""

    def __init__(self, commands: List[Command], plan_service: PlanService, repository: Repository):
        assert not any(c.subcommand is None for c in commands), "Command missing subcommand attribute"
        self._plan_service = plan_service
        self._repository = repository

        self._parser = argparse.ArgumentParser(prog='git-plan', description='A better workflow for git.')
        self._parser.add_argument('subcommand', type=str, nargs='?', help='The subcommand to run')
        self._parser.add_argument('--version', dest='version', action='store_true')
        subparsers = self._parser.add_subparsers(dest='subcommand')

        for command in commands:
            command.register_subparser(subparsers)
            command.set_cli(self)

        self._commands: Dict[str, Command] = {c.subcommand: c for c in commands}

    def parse(self, args: List[str]):
        """Parse command line arguments and invoke the correct subcommand

        Args:
            args:   Command-line arguments
        """
        parsed_args = self._parse_args(args)
        if parsed_args.version:
            self.version()
            return

        if not parsed_args.subcommand:
            if self._plan_service.has_commits(self._repository):
                parsed_args.subcommand = "list"
            else:
                parsed_args.subcommand = "add"

        try:
            self.invoke(**vars(parsed_args))  # Convert to dict
        except CommandNotFound:
            self.help()

    def invoke(self, subcommand: str, **kwargs):
        """Invoke the requested subcommand and pass kwargs to it

        Args:
            subcommand:     The requested subcommand
            kwargs:         The arguments for subcommand, parsed by its subparser
        """
        command = self.get_command(subcommand)
        command.run(kwargs)

    def get_command(self, subcommand: str) -> Command:
        """Return the requested command as a singleton

        Args:
            subcommand:     The requested subcommand

        Returns:
            The corresponding instance of Command
        """
        command = self._commands.get(subcommand)

        if not command:
            raise CommandNotFound()

        return command

    @staticmethod
    def version():
        """Print the version"""
        version = pkg_resources.require('git_plan')[0]
        print(version)

    def help(self):
        """Print the help"""
        self._parser.print_help()

    def _parse_args(self, args: List[str]) -> Namespace:
        """Main entrypoint for parsing arguments.

        Args:
            args:   The commandline args to parse
        """
        parsed_args = self._parser.parse_args(args)

        return parsed_args



"""Configuration

Author: Rory Byrne <rory@rory.bio>
"""
from pathlib import Path

from git_plan import constants


class Settings(dict):
    """Responsible for loading all settings from disk/defaults before starting the app

    Here, user-defined settings like custom templates can be loaded.
    """

    # Builders ####################################################################################

    @staticmethod
    def _from_project() -> dict:
        """Attempt to load a config.yaml from the local git repository"""
        return {}

    @staticmethod
    def _from_local() -> dict:
        """Attempt to load a config.yaml from a default directory"""
        return {}

    @staticmethod
    def _default() -> dict:
        """Return the default settings"""
        return constants.DEFAULT_SETTINGS

    @staticmethod
    def load() -> "Settings":
        """Load the settings"""
        settings = Settings._default()
        local_settings = Settings._from_local()
        project_settings = Settings._from_project()

        local_settings.update(project_settings)  # Roll project into local settings
        settings.update(local_settings)  # Roll both into the default settings

        cwd = Path.cwd()
        settings["working_dir"] = cwd

        return Settings(settings)



"""Repository model

Author: Rory Byrne <rory@rory.bio>
"""
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from git_plan.exceptions import NotAGitRepository
from git_plan.util.git import get_repository_root


@dataclass
class Repository:
    """A git repository"""
    root_dir: Path

    @property
    def plan_dir(self) -> Path:
        """This repo's .plan/ directory"""
        return Path(self.root_dir) / '.plan'

    @property
    def plan_files_dir(self) -> Path:
        """The directory where this repo's plans are stored"""
        return self.plan_dir / "plans"

    def is_initialized(self) -> bool:
        """Returns true if the repo has a plan directory"""
        return self.plan_dir.exists() and self.plan_files_dir.exists()

    @classmethod
    def from_working_dir(cls, working_dir: Path) -> Optional["Repository"]:
        """Constructs a Repository if in a git repository, otherwise returns None"""
        try:
            repo_root = get_repository_root(working_dir)
            return cls(repo_root)
        except NotAGitRepository:
            return None



"""Git utils

Author: Rory Byrne <rory@rory.bio>
"""
from pathlib import Path

from git_plan.exceptions import NotAGitRepository

GIT_DIR = '.git'


def get_repository_root(directory: Path):
    """Walks up the directory tree to find the root git dir

    Note:
        Technically, this will return true on a non-git repository if it has a .git/ directory
        ...But who would ever do that?
    """
    prev, directory = None, Path(directory).resolve()
    while prev != directory:
        if (directory / GIT_DIR).exists():
            return directory.resolve()
        prev, directory = directory, directory.parent

    raise NotAGitRepository()



"""List command

@author Rory Byrne <rory@rory.bio>
"""
from argparse import ArgumentParser
from typing import Any

from git_plan.cli.commands.command import Command
from git_plan.exceptions import NotAGitRepository
from git_plan.service.git import GitService
from git_plan.service.plan import PlanService
from git_plan.util.decorators import requires_initialized, requires_git_repository


@requires_initialized
@requires_git_repository
class List(Command):
    """List commits."""

    subcommand = 'list'

    def __init__(self, plan_service: PlanService, git_service: GitService, **kwargs):
        super().__init__(**kwargs)
        assert plan_service, "Plan service not injected"
        self._plan_service = plan_service
        self._git = git_service

    def command(self, *, long: bool = False, branch: bool = None, **kwargs):  # pylint: disable=arguments-differ
        """List the planned commits"""
        if not self._repository:
            raise NotAGitRepository()

        if branch:
            filter_branch = self._git.get_current_branch()
        else:
            filter_branch = None

        branch_display = filter_branch if filter_branch else "all branches"
        self._ui.print(f"Plans for [bold]{branch_display}[/bold]\n")
        commits = self._plan_service.get_commits(self._repository, branch=branch)

        if len(commits) == 0:
            if branch:
                self._ui.bold("No commit plans on the current branch.")
            else:
                self._ui.bold("No commit plans.")

            return

        commits = sorted(commits, key=lambda c: c.id)
        self._ui.render_commits(commits, headline_only=not long)

    def register_subparser(self, subparsers: Any):
        parser: ArgumentParser = subparsers.add_parser(List.subcommand, help='List existing commit plans.')
        parser.add_argument('-l', '--long', dest='long', action='store_true')
        parser.add_argument(
            '-b',
            '--branch',
            dest='branch',
            action='store_true',
            help='Show plans for the current branch'
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['git_plan',
 'git_plan.cli',
 'git_plan.cli.commands',
 'git_plan.model',
 'git_plan.service',
 'git_plan.util']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=4.2.1,<5.0.0',
 'dependency-injector[yaml]>=4.31.1,<5.0.0',
 'humanize>=3.3.0,<4.0.0',
 'inquirer>=2.7.0,<3.0.0',
 'rich>=9.13.0,<10.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['git-plan = git_plan.__cli__:main',
                     'gp = git_plan.__cli__:main']}

setup_kwargs = {
    'name': 'git-plan',
    'version': '0.2.4',
    'description': 'A better personal workflow for git',
    'long_description': None,
    'author': 'Rory Byrne',
    'author_email': 'rory@rory.bio',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/synek/git-plan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
